module AST where

import Data.Maybe (fromJust)
import Data.List (nub,elemIndex,sort,sortBy,(\\),intersect)
import ParseTree
import Debug.Trace
{-
  Abstract Syntax Tree
-}
-- a :: type for variables, b :: type for function names
type AST a b = [ ARule a b ]
type RuleID = Int
data ARule a b = ARule RuleID b [TTree a b] [TTree a b] [WRule a b]
   -- Left Patterns, Right Patterns, Where-Clauses

instance Eq (ARule a b) where
    (==) (ARule rid _ _ _ _)  (ARule rid' _ _ _ _) = rid == rid'

data WRule a b = WRule RuleID [a] b [a]
data TTree a b = F b [TTree a b]      -- Functions
               | C b [TTree a b]      -- Constructors
               | V a                       -- Variable
               deriving Eq
implode :: [a] -> [[a]] -> [a]
implode _ []  = []
implode _ [x] = x
implode m (x:y:xs) = x++(m++(implode m (y:xs)))

data SymbolName = SOrg String
--                | SCmp SymbolName [Int]
--                | STpl SymbolName [Int]
                | STpl SymbolName
                | SCmp SymbolName
                | SInv SymbolName
                | SNrm SymbolName [Int]
                | SSys String
                  deriving (Eq, Ord)

instance Show SymbolName where
    show (SOrg s) = s
    show (SCmp s) = "@"++show s
    show (STpl s) = "*"++show s
    show (SNrm s idxs) = show s ++ show idxs
    show (SInv s) = "!"++show s
    show (SSys s) = "#"++s

-- instance Ord SymbolName where
--     compare (SOrg s) (SOrg s') = compare s s'
--     compare (SCmp s is) (SCmp s' is') 
--         | s < s' = LT
--         | s > s' = GT
--         | True   = compare is is'
--     compare (STpl s is) (STpl s' is') 
--         | s < s' = LT
--         | s > s' = GT
--         | True   = compare is is'
--     compare (SNrm s is) (SNrm s' is') 
--         | s < s' = LT
--         | s > s' = GT
--         | 
--     compare (SInv s) (SInv s') = compare s s'
--     compare (SSys s) (SSys s') = compare s s'
--     compare (SOrg _) _ = LT
--     compare _ (SOrg _) = GT
--     compare (SCmp _ _) _ = LT
--     compare _ (SCmp _ _) = GT
--     compare (STpl _ _) _ = LT
--     compare _ (STpl _ _) = GT
--     compare (SSys _) _   = LT
--     compare _ (SSys _)   = GT


instance (Show a, Show b) => Show (ARule a b) where
    showList xs = \s->showList' (sortBy compRID xs) s
        where
          compRID (ARule i _ _ _ _) (ARule j _ _ _ _) = compare i j
          showList' [] = id
          showList' (ARule id fname lpats rpats wdecls:rest) =         
             (\s-> "Rule "++show id++":"++"\n"++
                  show fname++"("++implode "," (map show lpats)++")"++ " = " ++
                  showVsL rpats++ "\n" ++
                  (if null wdecls then [] else  "  where\n" ++ showWhere wdecls) ++"\n"++ s).showList rest
                where showWhere wdecls = 
                          let ls = map (\(WRule id lvs fname rvs)->showVsL lvs) wdecls
                              rs = map (\(WRule id lvs fname rvs)->show fname++showVsR rvs) wdecls
                              ml = maximum (map length ls)
                              ss = zipWith (\l r->"    "++fixSp ml l++" = "++ r++"\n") ls rs
                          in concat ss
                      showVsR vs = "("++implode "," (map show vs)++")"
                      showVsL [x] = show x
                      showVsL vs = showVsR vs
                      fixSp ml l = l ++ replicate (ml - length l) ' ' 


instance (Show a, Show b) => Show (TTree a b) where 
    show (F fname ls) = show fname++"("++implode "," (map show ls)++")"
    show (C cname ls) | null ls   = show cname
                      | otherwise = show cname++"("++implode "," (map show ls)++")"
    show (V vname)    = show vname

data AVariable b = VFromP Int
                 | VFromW b RuleID Int
                 | VFromF b Int
                 | VAny

instance Show a => Show (AVariable a) where
    show (VFromP i) = "_"++show i
    show (VFromW f rid i) = "_"++show f++"_"++show rid++"_"++show i
    show (VFromF f i)     = "_"++show f++"_"++show i
    show (VAny)           = "_*"

instance Eq a => Eq (AVariable a) where
    (==) (VFromP i) (VFromP i') = i==i'
    (==) (VFromW f rid i) (VFromW f' rid' i') = f==f' && rid==rid' && i==i'
    (==) (VFromF f i) (VFromF f' i') = f==f' && i==i'
    (==) (VAny) (VAny) = True
    (==) _ _ = False

instance Ord a => Ord (AVariable a) where
    compare (VFromP i) (VFromP i') = compare i i'
    compare (VFromW f rid i) (VFromW f' rid' i') 
            | f < f' = LT
            | f > f' = GT
            | i < i' = LT
            | i > i' = GT
            | True   = compare rid rid'
    compare (VFromF f i) (VFromF f' i') 
            | f < f' = LT
            | f > f' = GT
            | True   = compare i i'
    compare (VFromP _) (VFromW _ _ _) = LT
    compare (VFromP _) (VFromF _ _)   = LT
    compare (VFromP _) (VAny)         = LT
    compare (VFromW _ _ _) (VFromF _ _) = LT
    compare (VFromW _ _ _) (VAny)       = LT
    compare (VFromF _ _) (VAny)         = LT
    compare _ _ = GT

removeUnusedRules g rs = filter (\(ARule _ f _ _ _) -> f `elem` (g:usedFunctions)) rs
    where usedFunctions = nub $ concatMap takeWRSymb rs
          takeWRSymb (ARule _ _ _ _ wdecls) = map (\(WRule _ _ f _)->f) wdecls

prettyVariables f rs = map prettyvars $ map removeIDCall $ filter (\(ARule _ f _ _ _)-> f /= idSymb) (removeUnusedRules f rs)
    where
      removeIDCall (ARule rid f ps es wdecls) = ARule rid f (map conv ps) (map conv es) wdecls'
          where bind = nub $ concatMap (\(WRule _ vs _ vs')->zip vs vs') $ filter (\(WRule _ _ f  _)->f==idSymb) wdecls
                conv (C c ts) = C c (map conv ts)
                conv (F f ts) = F f (map conv ts)
                conv (V v)    = case lookup v bind of 
                                  Just v' -> V v'
                                  Nothing -> V v
                wdecls' = filter (\(WRule _ _ f _)->f/=idSymb) wdecls                
      prettyvars (ARule rid f ps es wdecls) = ARule rid f (map conv ps) (map conv es) wdecls'
              where vars (C c ts) = concatMap vars ts
                    vars (F f ts) = concatMap vars ts
                    vars (V v)    = [v]
                    variablesU = nub$concatMap vars es++concatMap (\(WRule i vs _ vs')-> vs') wdecls 
                    variablesP = variablesU `intersect` (nub$concatMap vars ps)
                    variablesW = variablesU `intersect` (nub$concatMap (\(WRule i vs _ vs')-> vs) wdecls)
                    conv (C c ts) = C c (map conv ts)
                    conv (F f ts) = F f (map conv ts)
                    conv (V v)    = V$convV v
                    convV v = case elemIndex v variablesP of
                                   Just i  -> SOrg $ "x"++show i
                                   Nothing -> 
                                       case elemIndex v variablesW of 
                                         Just i -> SOrg $ "t"++show i
                                         Nothing -> SOrg $ "_"
                    wdecls' = map (\(WRule i vs f vs')->WRule i (map convV vs) f (map convV vs')) wdecls
                              
-- Symbol for ID
idSymb = SSys "I" 
  
maximumRID rules  = foldr max 0 [ rid | ARule rid _ _ _ _ <- rules ]
maximumWID wdecls = foldr max 0 [ wid | WRule wid _ _ _ <- wdecls ]
rulesOf    f rules = [ ARule i f' ps es ws | ARule i f' ps es ws <- rules, f==f'] 
rulesNotOf f rules = [ ARule i f' ps es ws | ARule i f' ps es ws <- rules, f/=f'] 
        
variables (F _ ts) = concatMap variables ts
variables (C _ ts) = concatMap variables ts
variables (V v)    = [v]

normalizeProgram f rules = 
    fix [] (map (\(ARule id f ps es ws)->(ARule id f ps es ws, [0..(length es-1)]) ) frules) 1 []
    where
      frules = rulesOf f rules
      fix rls [] cnt proced = rls
      fix rls ((r,idxs):reqs) cnt proced =
          if (r,idxs) `elem` proced then
              fix rls reqs cnt proced 
          else
              let (r',reqf') = norm r idxs cnt 
                  reqs' = concatMap (\(f,idxs) -> map (\r->(r,idxs))$rulesOf f rules) reqf'
              in fix (r':rls) (reqs'++reqs) (cnt+1) ((r,idxs):proced)
      norm (ARule rid f ps es wdecls) idxs cnt = (ARule cnt (SNrm f idxs) ps es' wdecls', require)
          where   
            swdecls = sortBy (\(WRule i _ _ _) (WRule j _ _ _) -> compare i j) wdecls
            wids     = [ i | WRule i _ _ _ <- swdecls ]
            avars = nub $ concat [ variables (es !! i) | i <- idxs ]
            jidxss = map (\wid ->nub$sort$concatMap (\v->
                                      case v of VFromW _ i j | i == wid -> [j]
                                                _ -> []) avars) wids
            wi_js = zip wids jidxss
            es' = map convV $ map (es !!) idxs            
            convV (C c ts) = C c (map convV ts)
            convV (F f ts) = F f (map convV ts)
            convV (V (VFromW f i j))    = V $ convVar (fromJust$lookup i wi_js) (VFromW f i j)
            convV (V v) = trace ("V::"++show v) $ V v
            convVar [0]  (VFromW f i j) | f == idSymb = VFromW idSymb i j 
            convVar idxs (VFromW f i j)               = VFromW (SNrm f idxs) i (fromJust$elemIndex j idxs)
            convVar _ v = trace ("V:"++show v) v
            wdecls' = zipWith ((\(WRule i vs f vs') jdxs ->
                                    if f == idSymb && jdxs == [0] then
                                        WRule i vs f vs'
                                    else
                                        WRule i (map (convVar jdxs)$map (vs!!) jdxs) (SNrm f jdxs) vs'
                               )) swdecls jidxss
            require = concatMap (\(WRule _ _ f _) -> case f of
                                                       SNrm f jdxs -> [(f,jdxs)]
                                                       _ ->           [] ) wdecls'
            
          

removeUnusedWhere rules = 
    map rmunused rules
    where rmunused (ARule rid f ps es wdecls) = ARule rid f ps es wdecls'
              where wdecls' = filter (\(WRule wid vs f vs') -> any (\v->any (appear v) es) vs) wdecls
                    appear v (C _ ts) = any (appear v) ts
                    appear v (F _ ts) = any (appear v) ts
                    appear v (V v')   = v==v'

introduceIdentity conv1 = 
    idDef ++ map introduceID rules
    where rules = rulesNotOf idSymb conv1 
          len   = maximumRID conv1+1
          conts = nub $ concatMap (\(ARule _ _ ps es wdecls)->
                                       concatMap allC ps++concatMap allC es) rules
          allC (C c ts) = (c,length ts):concatMap allC ts
          allC (F f ts) = concatMap allC ts
          allC (V _)    = []
          idDef = zipWith mkRule [len..(len+length conts-1)] conts
          mkRule id (c,len) = ARule id idSymb [(C c $ map V vs)] [(C c $ map V vs')] wdecls
              where vs  = map VFromP [0..len-1]
                    vs' = map (\i->VFromW idSymb i 0) [0..len-1]
                    wdecls = zipWith (\i (v,v')->WRule i [v'] idSymb [v]) [0..len-1] (zip vs vs')
          introduceID (ARule rid f ps es wdecls) = ARule rid f ps (map conv es) (wdecls++nwds)
              where wlen = maximumWID wdecls+1
                    pvars (C _ ts) = concatMap pvars ts
                    pvars (F _ ts) = concatMap pvars ts
                    pvars (V (VFromP i)) = [VFromP i]
                    pvars (V _) = []
                    rhspVars   = nub $ concatMap (\(WRule _ _ _ vs)->
                                                      filter (\v->case v of
                                                                    VFromP i -> True
                                                                    _ -> False) vs) wdecls
                    pVariables = (nub $ concatMap pvars ps) \\ rhspVars
                    nwids = [wlen..(wlen+length pVariables-1)]
                    vs = map (\i -> VFromW idSymb i 0) nwids
                    nwds = zipWith3 (\i v v' -> WRule i [v] idSymb [v']) nwids vs pVariables
                    vp = zip pVariables vs
                    conv (C c ts) = C c (map conv ts)
                    conv (F f ts) = F f (map conv ts)
                    conv (V v)    = case lookup v vp of
                                      Just v' -> V v'
                                      Nothing -> V v

--
-- f(C(a,b),c) = ( C(s,t), x, b ) where (s,t) = f(a)
--                                      (x,y) = f(b)
-- f(C(_0,_1),_2) = ( C(_f_0_0,_f_0_1), _f_1_0, _1 ) where ...
-- 


convertToAST :: [Rule] -> AST (AVariable SymbolName) SymbolName
convertToAST rs = introduceIdentity$zipWith (\i r->convRule i (introduceWhere r)) [0..(length rs-1)] rs
          

introduceWhere :: Rule -> Rule
introduceWhere (Rule (fname, PTpl pts, ETpl exs, wdecls)) =
    Rule (fname, PTpl pts, ETpl (map conv exs), wdecls')
    where fcalls (ECon _ exs) = concatMap fcalls exs
          fcalls (EFun f exs) = [EFun f exs]
          fcalls (EVar v)     = []
          fcs = nub$concatMap fcalls exs
          assignVarName fc = "_"++(show $ elemIndex fc fcs)
          conv (ECon c exs) = ECon c (map conv exs)
          conv (EFun f exs) = EVar (assignVarName (EFun f exs))
          conv (EVar v)     = EVar v
          wdecls' = wdecls ++ map (\fc -> WDecl (PVar (assignVarName fc), fc)) fcs 

convRule :: Int -> Rule -> ARule (AVariable SymbolName) SymbolName
convRule i (Rule (fname,PTpl pts,ETpl exs,wdecls)) = 
    let pbind = mkBindP pts
        wbind = mkBindW wdecls
        bind  = pbind++wbind
    in ARule i
         (SOrg fname)
         (map (subVar bind.convP) pts) 
         (map (subVar bind.convE) exs)
         (zipWith (\w i->convW bind i w) wdecls [0..])
    where 
      mkBindP pts    = zip (concatMap varsP pts) (map VFromP [0..])
      mkBindW wdecls = concat$zipWith (\i w->mkBW i w) [0..] wdecls
      mkBW i (WDecl (p,EFun f _)) = zip (varsP p) (map (\j -> VFromW (SOrg f) i j) [0..])
      varsP (PVar s)    = [s]
      varsP (PCon _ ps) = concatMap varsP ps
      varsP (PTpl ps)   = concatMap varsP ps

subVar bind t = subVar' t 
    where subVar' (F f ts) = F f (map subVar' ts)
          subVar' (C f ts) = C f (map subVar' ts)
          subVar' (V v)    = case lookup v bind of
                               Just v' -> V v'

convP (PVar s)    = V s
convP (PCon c ps) = C (SOrg c) (map convP ps)

convertExp (ECon c ex) = C (SOrg c) (map convertExp ex)
convertExp (EFun f ex) = F (SOrg f) (map convertExp ex)

convE (ECon c ex) = C (SOrg c) (map convE ex)
convE (EFun f ex) = F (SOrg f) (map convE ex)
convE (EVar v)    = V v

convW bind i (WDecl (PVar v, EFun f es)) = 
    convW bind i (WDecl (PTpl [PVar v], EFun f es))
convW bind i (WDecl (PTpl ps, EFun f es)) = 
    WRule i (map (\(PVar v)->fromJust$lookup v bind) ps) (SOrg f)
            (map (\(EVar v)->fromJust$lookup v bind) es)

